@@ -3,11 +3,12 @@ package LCS::BV;
 use 5.010001;
 use strict;
 use warnings;
-our $VERSION = '0.08';
+our $VERSION = '0.10';
 #use utf8;
 
 our $width = int 0.999+log(~0)/log(2);
 
+use integer;
 no warnings 'portable'; # for 0xffffffffffffffff
 
 sub new {
@@ -24,7 +25,7 @@ sub new {
 sub LLCS {
   my ($self,$a,$b) = @_;
 
-  use integer;
+  #use integer;
   #no warnings 'portable'; # for 0xffffffffffffffff
 
   # TODO: maybe faster, if we have fewer expensive iterations
@@ -36,7 +37,6 @@ sub LLCS {
 
   my ($amin, $amax, $bmin, $bmax) = (0, $#$a, 0, $#$b);
 
-  #if (1) {
   while ($amin <= $amax and $bmin <= $bmax and $a->[$amin] eq $b->[$bmin]) {
     $amin++;
     $bmin++;
@@ -45,12 +45,68 @@ sub LLCS {
     $amax--;
     $bmax--;
   }
-  #}
 
-  my $positions;
+  #my %positions;
 
-  if (1 && $amax < $width ) {
-    $positions->{$a->[$_]} |= 1 << ($_ % $width) for $amin..$amax;
+  if ($amax < $width ) {
+    my %positions;
+    #$positions{$a->[$_]}  |= 1 << ($_ % $width) for $amin..$amax;
+    for ($amin..$amax) { $positions{$a->[$_]} |= 1 << ($_ % $width); }
+
+    my $v = ~0;
+    my ($p,$u);
+
+    for ($bmin..$bmax) {
+      $p = $positions{$b->[$_]} // 0;
+      $u = $v & $p;
+      $v = ($v + $u) | ($v - $u);
+    }
+    return $amin + _count_bits(~$v) + $#$a - $amax;
+  }
+  else {
+    my %positions;
+    #$positions{$a->[$_]}->[$_ / $width] |= 1 << ($_ % $width) for $amin..$amax;
+    for ($amin..$amax) { $positions{$a->[$_]}->[$_ / $width] |= 1 << ($_ % $width); }
+
+    my $S;
+    my @Vs = ();      # $Vs->[$k] = bits;
+
+    my ($p, $u, $carry);
+
+    my $kmax = ($amax+1) / $width;
+    $kmax++ if (($amax+1) % $width);
+
+    for (my $k=0; $k < $kmax; $k++ ) { $Vs[$k] = ~0; }
+
+    for my $j ($bmin..$bmax) {
+      $carry = 0;
+      for (my $k=0; $k < $kmax; $k++ ) {
+        $S = $Vs[$k];
+        $p = $positions{$b->[$j]}->[$k] // 0;
+        $u = $S & $p;             # [Hyy04]
+        $Vs[$k] = ($S + $u + $carry) | ($S - $u);
+        $carry = (($S & $u) | (($S | $u) & ~($S + $u + $carry))) >> ($width-1) & 1;
+      }
+    }
+
+    my $bitcount = 0;
+
+    for my $k ( @Vs ) {
+      $bitcount += _count_bits(~$k);
+    }
+    return $amin + $bitcount + $#$a - $amax;
+  }
+}
+
+sub LLCS_prepared {
+  my ($self,$positions,$b) = @_;
+
+  #if ( ref($a) eq 'HASH' )    # $a contains $positions
+
+  my ($bmin, $bmax) = (0, $#$b);
+  my ($amin, $amax) = (0, $self->{'length_a'} - 1);
+
+  if ($amax < $width ) {
 
     my $v = ~0;
     my ($p,$u);
@@ -60,41 +116,37 @@ sub LLCS {
       $u = $v & $p;
       $v = ($v + $u) | ($v - $u);
     }
-    return $amin + _count_bits(~$v) + scalar(@$a) - ($amax+1);
+    return _count_bits(~$v);
   }
   else {
-    $positions->{$a->[$_]}->[$_ / $width] |= 1 << ($_ % $width) for $amin..$amax;
 
     my $S;
-    my $Vs = [];      # $Vs->[$k] = bits;
+    my @Vs = ();      # $Vs->[$k] = bits;
 
     my ($p, $u, $carry);
 
     my $kmax = ($amax+1) / $width;
     $kmax++ if (($amax+1) % $width);
 
-    for (my $k=0; $k < $kmax; $k++ ) { $Vs->[$k] = ~0; }
+    for (my $k=0; $k < $kmax; $k++ ) { $Vs[$k] = ~0; }
 
     for my $j ($bmin..$bmax) {
       $carry = 0;
       for (my $k=0; $k < $kmax; $k++ ) {
-        #$S = (exists($Vs->[$k])) ? $Vs->[$k] : ~0;
-        $S = $Vs->[$k];
+        $S = $Vs[$k];
         $p = $positions->{$b->[$j]}->[$k] // 0;
         $u = $S & $p;             # [Hyy04]
-        $Vs->[$k] = ($S + $u + $carry) | ($S - $u);
+        $Vs[$k] = ($S + $u + $carry) | ($S - $u);
         $carry = (($S & $u) | (($S | $u) & ~($S + $u + $carry))) >> ($width-1) & 1;
       }
     }
 
     my $bitcount = 0;
 
-    if (@$Vs) {
-      for my $k ( @{$Vs} ) {
-        $bitcount += _count_bits(~$k);
-      }
+    for my $k ( @Vs ) {
+      $bitcount += _count_bits(~$k);
     }
-    return $amin + $bitcount + scalar(@$a) - ($amax+1);
+    return $bitcount;
   }
 }
 
@@ -102,7 +154,7 @@ sub LLCS {
 sub LCS {
   my ($self, $a, $b) = @_;
 
-  use integer;
+  #use integer;
   #no warnings 'portable'; # for 0xffffffffffffffff
 
   my ($amin, $amax, $bmin, $bmax) = (0, $#$a, 0, $#$b);
@@ -116,24 +168,23 @@ sub LCS {
     $bmax--;
   }
 
-
-  my $positions;
+  my %positions;
   my @lcs;
 
   if ($amax < $width ) {
-    $positions->{$a->[$_]} |= 1 << ($_ % $width) for $amin..$amax;
+    $positions{$a->[$_]} |= 1 << ($_ % $width) for $amin..$amax;
 
     my $S = ~0;
+    my @Vs = (~0);
 
-    my $Vs = [];
     my ($y,$u);
 
     # outer loop
     for my $j ($bmin..$bmax) {
-      $y = $positions->{$b->[$j]} // 0;
+      $y = $positions{$b->[$j]} // 0;
       $u = $S & $y;               # [Hyy04]
       $S = ($S + $u) | ($S - $u); # [Hyy04]
-      $Vs->[$j] = $S;
+      $Vs[$j] = $S;
     }
 
     # recover alignment
@@ -141,14 +192,12 @@ sub LCS {
     my $j = $bmax;
 
     while ($i >= $amin && $j >= $bmin) {
-      if ($Vs->[$j] & (1<<$i)) {
+      if ($Vs[$j] & (1<<$i)) {
         $i--;
       }
       else {
         unless (
-           $j
-           && exists $Vs->[$j-1]
-           && ~$Vs->[$j-1] & (1<<$i)
+           $j && ~$Vs[$j-1] & (1<<$i)
         ) {
            unshift @lcs, [$i,$j];
            $i--;
@@ -158,10 +207,10 @@ sub LCS {
     }
   }
   else {
-    $positions->{$a->[$_]}->[$_ / $width] |= 1 << ($_ % $width) for $amin..$amax;
+    $positions{$a->[$_]}->[$_ / $width] |= 1 << ($_ % $width) for $amin..$amax;
 
     my $S;
-    my $Vs = [];
+    my @Vs = ([~0]);
     my ($y,$u,$carry);
 
     my $kmax = ($amax+1) / $width;
@@ -169,15 +218,15 @@ sub LCS {
 
     # outer loop
     for my $j ($bmin..$bmax) {
-      for (my $k=0; $k < $kmax; $k++ ) { $Vs->[$j]->[$k] = ~0; }
+      for (my $k=0; $k < $kmax; $k++ ) { $Vs[$j]->[$k] = ~0; }
       $carry = 0;
 
       for (my $k=0; $k < $kmax; $k++ ) {
-        $S = ($j > $bmin) ? $Vs->[$j-1]->[$k] : ~0;
-        $y = $positions->{$b->[$j]}->[$k] // 0;
+        $S = ($j > $bmin) ? $Vs[$j-1]->[$k] : ~0;
+        $y = $positions{$b->[$j]}->[$k] // 0;
         $u = $S & $y;             # [Hyy04]
 
-        $Vs->[$j]->[$k] = ($S + $u + $carry) | ($S - $u);
+        $Vs[$j]->[$k] = ($S + $u + $carry) | ($S - $u);
 
         $carry = (($S & $u) | (($S | $u) & ~($S + $u + $carry))) >> ($width-1) & 1;
       }
@@ -189,14 +238,12 @@ sub LCS {
 
     while ($i >= $amin && $j >= $bmin) {
       my $k = $i / $width;
-      if ($Vs->[$j]->[$k] & (1<<($i % $width))) {
+      if ($Vs[$j]->[$k] & (1<<($i % $width))) {
         $i--;
       }
       else {
         unless (
-           $j
-           && exists $Vs->[$j-1]->[$k]
-           && ~$Vs->[$j-1]->[$k] & (1<<($i % $width))
+           $j && ~$Vs[$j-1]->[$k] & (1<<($i % $width))
         ) {
            unshift @lcs, [$i,$j];
            $i--;
@@ -207,16 +254,32 @@ sub LCS {
   }
 
   return [
-    map([$_ => $_], 0 .. ($bmin-1)),
+    map([$_ => $_], 0 .. ($bmin-1)), ## no critic qw(BuiltinFunctions::RequireBlockMap)
     @lcs,
-    map([++$amax => $_], ($bmax+1) .. $#$b)
+    map([++$amax => $_], ($bmax+1) .. $#$b) ## no critic qw(BuiltinFunctions::RequireBlockMap)
   ];
 }
 
+sub prepare {
+  my ($self, $a) = @_;
+
+  $self->{'length_a'} = @{$a};
+
+  my $positions;
+
+  if ($#$a < $width ) {
+    $positions->{$a->[$_]} |= 1 << ($_ % $width) for 0..$#$a;
+  }
+  else {
+    $positions->{$a->[$_]}->[$_ / $width] |= 1 << ($_ % $width) for 0..$#$a;
+  }
+  return $positions;
+}
+
 sub _count_bits {
   my $v = shift;
 
-  use integer;
+  #use integer;
   #no warnings 'portable'; # for 0xffffffffffffffff
 
   if ($width == 64) {
@@ -244,6 +307,10 @@ sub _count_bits {
 
 __END__
 
+=pod
+
+=encoding UTF-8
+
 =head1 NAME
 
 LCS::BV - Bit Vector (BV) implementation of the
@@ -262,8 +329,14 @@ LCS::BV - Bit Vector (BV) implementation of the
 
   use LCS::BV;
 
-  $alg = LCS::BV->new;
-  @lcs = $alg->LCS(\@a,\@b);
+  $object = LCS::BV->new;
+  @lcs    = $object->LCS(\@a,\@b);
+
+  $llcs   = $object->LLCS(\@a,\@b);
+
+  $positions = $object->prepare(\@a);
+  $llcs      = $object->LLCS_prepared($positions,\@b);
+
 
 =head1 ABSTRACT
 
@@ -280,7 +353,7 @@ than the used wordsize of the hardware (32 or 64 bits).
 =item new()
 
 Creates a new object which maintains internal storage areas
-for the LCS computation.  Use one of these per concurrent
+for the LCS computation. Use one of these per concurrent
 LCS() call.
 
 =back
@@ -300,15 +373,92 @@ Finds a Longest Common Subsequence, taking two arrayrefs as method
 arguments. It returns an array reference of corresponding
 indices, which are represented by 2-element array refs.
 
+=item prepare(\@a)
+
+For comparison of the same sequence @a many times against different sequences @b
+it can be faster to prepare @a. It returns a hashref of positions.
+This works only on object mode.
+
+=item LLCS_prepared($positions,\@a)
+
+Return the length of a Longest Common Subsequence, with one side prepared.
+It returns an integer.
+
 =back
 
 =head2 EXPORT
 
 None by design.
 
+=head1 STABILITY
+
+Until release of version 1.00 the included methods, names of methods and their
+interfaces are subject to change.
+
+Beginning with version 1.00 the specification will be stable, i.e. not changed between
+major versions.
+
+=head1 REFERENCES
+
+H. Hyyroe. A Note on Bit-Parallel Alignment Computation.
+In M. Simanek and J. Holub, editors, Stringology, pages 79-87. Department
+of Computer Science and Engineering, Faculty of Electrical
+Engineering, Czech Technical University, 2004.
+
+=head1 SPEED COMPARISON
+
+    Intel Core i7-4770HQ Processor
+    Intel® SSE4.1, Intel® SSE4.2, Intel® AVX2
+    4 Cores, 8 Threads
+    2.20 - 3.40 GHz
+    6 MB Cache
+    16 GB DDR3 RAM
+
+LCS-BV/xt$ perl 50_diff_bench.t
+
+LCS: Algorithm::Diff, Algorithm::Diff::XS, LCS, LCS::BV
+
+LCS: [Chrerrplzon] [Choerephon]
+
+                Rate      LCS:LCS    AD:LCSidx AD:XS:LCSidx   LCS:BV:LCS
+LCS:LCS       9225/s           --         -72%         -87%         -88%
+AD:LCSidx    33185/s         260%           --         -53%         -58%
+AD:XS:LCSidx 70447/s         664%         112%           --         -12%
+LCS:BV:LCS   79644/s         763%         140%          13%           --
+
+LCS: [qw/a b d/ x 50], [qw/b a d c/ x 50]
+
+               Rate      LCS:LCS    AD:LCSidx   LCS:BV:LCS AD:XS:LCSidx
+LCS:LCS      49.5/s           --         -37%         -96%         -99%
+AD:LCSidx    79.0/s          60%           --         -94%         -98%
+LCS:BV:LCS   1255/s        2434%        1488%           --         -69%
+AD:XS:LCSidx 4073/s        8124%        5052%         224%           --
+
+LLCS: [Chrerrplzon] [Choerephon]
+
+                     Rate    LCS:LLCS AD:LCS_length AD:XS:LCS_length LCS:BV:LLCS
+LCS:LLCS          11270/s          --          -70%             -70%        -92%
+AD:LCS_length     37594/s        234%            --              -1%        -75%
+AD:XS:LCS_length  38059/s        238%            1%               --        -74%
+LCS:BV:LLCS      148945/s       1222%          296%             291%          --
+
+LLCS: [qw/a b d/ x 50], [qw/b a d c/ x 50]
+
+                   Rate     LCS:LLCS AD:LCS_length AD:XS:LCS_length  LCS:BV:LLCS
+LCS:LLCS         50.0/s           --          -37%             -37%         -98%
+AD:LCS_length    79.2/s          58%            --              -1%         -97%
+AD:XS:LCS_length 79.8/s          60%            1%               --         -97%
+LCS:BV:LLCS      2357/s        4614%         2874%            2853%           --
+
+=head1 SOURCE REPOSITORY
+
+L<http://github.com/wollmers/LCS-BV>
+
 =head1 SEE ALSO
 
 Algorithm::Diff
+LCS
+LCS::Tiny
 
 =head1 AUTHOR
 
@@ -322,7 +472,9 @@ Helmut Wollmersdorfer E<lt>helmut.wollmersdorfer@gmail.comE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 2014-2019 by Helmut Wollmersdorfer
+=encoding UTF-8
+
+Copyright 2014-2020 by Helmut Wollmersdorfer
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
